/*
  mmalloc - heap manager based on heavy use of virtual memory management.
  Copyright (C) 1998   Valery Shchedrin

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
  MA 02111-1307, USA
  
  Public Functions:
  
  void *mmalloc(size_t size);
  
    Allocates `size` bytes
    returns NULL if no free memory available
  
  void *mcalloc(size_t unit, size_t quantity);
  
    Allocates `quantity*unit` zeroed bytes via internal malloc call
  
  void *mrealloc(void *ptr, size_t size);
  
    Reallocates already allocated block `ptr`, if `ptr` is not valid block
    then it works as malloc. NULL is returned if no free memory available
  
  void *mrealloc_no_move(void *ptr, size_t size);
  
    Reallocates already allocated block `ptr`, if `ptr` is not valid block
    or if reallocation can't be done with shrinking/expanding already
    allocated block NULL is returned
  
  void mfree(void *ptr);
  
    Frees already allocated block, if `ptr` is incorrect one nothing will
    happen.
  
*/

#include "StdAfx.h"

#include "m_malloc.h"
#include "malloc.h"

typedef CRITICAL_SECTION mutex_t;

#if M_MULTITHREADED
# define mutex_init(x) ( InitializeCriticalSection(x) )
# define mutex_lock(x) ( EnterCriticalSection(x) )
# define mutex_unlock(x) ( LeaveCriticalSection(x) )
#else
# define mutex_init(x) ( (void) (x) )
# define mutex_unlock(x) ( (void) (x) )
# define mutex_lock(x) ( (void) (x) )
#endif

int __malloc_initialized = -1;
 /* -1 == uninitialized, 0 == initializing, 1 == initialized */

static mutex_t malloc_lock;

#ifndef NULL
#define NULL ((void*)0)
#endif

/* HUNK MANAGER */

typedef struct Hunk_s  Hunk_t;

#pragma pack(1)

struct Hunk_s { /* Hunked block - 8 byte overhead */
  int id; /* unique id */
  unsigned int total:12, used:12, size : 8;
  Hunk_t *next; /* next free in free_h */
};

#pragma pack()

static Hunk_t *free_h[HUNK_MAXSIZE+1]; /* free hash */
int total_h[HUNK_MAXSIZE+1]; /* Hunk_t's `total` member */

#define usagemap(h) (((unsigned char *)(h))+sizeof(Hunk_t))
#define hunk_ptr(h) (((char*)(h))+sizeof(Hunk_t)+ALIGN(DIV8(h->total+7)))
#define hunk(h)  ((Hunk_t*)(h))

/* hunk_alloc allocates <= HUNK_MAXSIZE blocks */
static void *hunk_alloc(int size)
{
  Hunk_t *p;
  unsigned long *cpl;
  int i, c;
  
  if (size >= HUNK_THRESHOLD) size = ALIGN(size);
  
  /* Look for already allocated hunkblocks */
  if ((p = free_h[size]) == NULL)
  {
    if ((p = (Hunk_t*)VirtualAlloc(NULL,HUNK_MSIZE,
		MEM_COMMIT,PAGE_EXECUTE_READWRITE)) == NULL)
      return NULL;
    memset(p,0,HUNK_MSIZE);
    p->id = HUNK_ID;
    p->total = total_h[size];
    /* p->used = 0; */
    p->size = size;
    /* p->next = (Hunk_t*)NULL; */
    /* memset(usagemap(p), 0, bound); */
    free_h[size] = p;
  }

  /* Locate free point in usagemap */
  for (cpl=(unsigned long*)usagemap(p);*cpl==0xFFFFFFFF;cpl++);
  i = ((unsigned char *)cpl) - usagemap(p);
  if (*(unsigned short*)cpl != 0xFFFF) {
    if (*(unsigned char*)cpl == 0xFF) {
      c = *(int*)(((unsigned char *)cpl)+1); i++;
      } else  c = *(int*)(unsigned char *)cpl;
  } else {
    i+=2; c = *(((unsigned char *)cpl)+2);
    if (c == 0xFF) { c = *(int*)(((unsigned char *)cpl)+3); i++; }
  }
  EMUL8(i);
  if ((c & 0xF) == 0xF) { c >>= 4; i+=4; }
  if ((c & 0x3) == 0x3) { c >>= 2; i+=2; }
  if (c & 1) i++;
  
  usagemap(p)[DIV8(i)] |= (1 << (i & 7)); /* set bit */
  
  /* Increment counter and update hashes */
  if (++p->used == p->total)
  {
    free_h[p->size] = p->next;
    p->next = NULL;
  }
  return hunk_ptr(p)+i*p->size;
}

/* hunk_free frees blocks allocated by hunk_alloc */
static void hunk_free(char *ptr)
{
  unsigned char *up;
  int i, v;
  Hunk_t *h;
  
  if (!ptr) return;
  
  h = (Hunk_t*)PAGE_DOWNALIGNP(ptr);
  
  /* Validate `ptr` */
  if (h->id != HUNK_ID) return;
  v = ptr - hunk_ptr(h);
  i = v / h->size;
  if (v % h->size != 0 || i < 0 || i >= (int)h->total) return;
  
  /* Update `usagemap` */
  up = &(usagemap(h)[DIV8(i)]);
  i = 1 << (i&7);
  if (!(*up & i)) return;
  *up ^= i;
  
  /* Update hunk counters */
  if (h->used == h->total)
  {
    if (--h->used)
    { /* insert into free_h */
      h->next = free_h[h->size];
      free_h[h->size] = h;
    } /* else - it will be unmapped */
  }
  else
  {
    if (!--h->used)
    { /* delete from free_h - will be bl_freed*/
      Hunk_t *p, *pp;
      for (p=free_h[h->size],pp=NULL;p!=h;pp=p,p=p->next);
      if (!pp)
	free_h[h->size] = p->next;
      else
	pp->next = p->next;
    }
  }
  
  /* Unmap empty Hunk_t */
  if (!h->used) VirtualFree(h,HUNK_MSIZE,MEM_RELEASE);
}

/* BLOCK MANAGER */

typedef struct Block_s Block_t;

#pragma pack(1)

struct Block_s /* 32-bytes long control structure (if 4-byte aligned) */
{
  char *ptr;                        /* pointer to related data */
  Block_t *next;                    /* next in free_mem list */
  Block_t *l_free_mem, *r_free_mem; /* left & right subtrees of <free_mem> */
  Block_t *l_ptrs, *r_ptrs;         /* left & right subtrees of <ptrs> */
  size_t size;                      /* size - divided by align */

  /* packed 4-byte attributes */
/* { */
  char bal_free_mem;   /* balance of <free_mem> subtree */
  char bal_ptrs;       /* balance of <ptrs> subtree */
  BYTE used;   /* used/free state of the block */
  BYTE broken; /* 1 if previous block can't be merged with it */
/* } */
};

#pragma pack()

static Block_t *bl_last; /* last mmapped block */

#define bl_get() hunk_alloc(sizeof(Block_t))
#define bl_rel(p) hunk_free((char*)p)

/* like C++ templates ;-) */
#include "avlmacro.h"

#define FREE_MEM_COMPARE(i,a,b) { i = (a)->size - (b)->size; }
#define PTRS_COMPARE(i,a,b) { i = (a)->ptr - (b)->ptr; }

Avl_Tree(free_mem,Block_t,free_mem,FREE_MEM_COMPARE)
Avl_Tree(ptrs,Block_t,ptrs,PTRS_COMPARE)

#define free_mem_root Avl_Root(Block_t, free_mem)
#define ptrs_root Avl_Root(Block_t, ptrs)

/* pp is freed block */
#define FREE_MEM_DEL_BLOCK(pp) \
{ \
  for (p = free_mem_root;;) \
    if (p->size > pp->size) p = p->l_free_mem; \
    else if (p->size < pp->size) p = p->r_free_mem; \
    else break; \
  if (p == pp) \
  { \
    if (pp->next) free_mem_replace(pp->next); \
    else free_mem_del(pp); \
  } \
  else \
  { \
    for (;p->next != pp; p = p->next); \
    p->next = pp->next; \
  } \
}

#define FREE_MEM_INS_BLOCK(pp) \
{ \
  if ((p = free_mem_ins(pp)) != NULL)\
  {\
    pp->next = p->next;\
    p->next = pp;\
  }\
  else pp->next = NULL; \
}

/* `b` is current block, `pp` is next block */
#define COMBINE_BLOCKS(b,pp) \
{\
  ptrs_del(pp); \
  b->size += pp->size; \
  if (pp == bl_last) bl_last = b; \
  bl_rel(pp); \
}

/* initializes new block b */
#define INIT_BLOCK(b, pppp, sz) \
{ \
  memset(b, 0, sizeof(Block_t)); \
  b->ptr  = pppp; \
  b->size = sz; \
  ptrs_ins(b); \
  FREE_MEM_INS_BLOCK(b); \
}

/* `b` is current block, `sz` its new size */
/* block `b` will be splitted to one busy & one free block */
#define SPLIT_BLOCK(b,sz) \
{\
  Block_t *bt; \
  bt = (Block_t*)bl_get(); \
  INIT_BLOCK(bt, b->ptr + sz, b->size - sz); \
  b->size = sz; \
  if (bl_last == b) bl_last = bt; \
  bl_uncommit(bt);\
}

/* `b` is current block, `pp` is next free block, `sz` is needed size */
#define SHRINK_BLOCK(b,pp,sz) \
{\
  FREE_MEM_DEL_BLOCK(pp); \
  pp->ptr = b->ptr + sz; \
  pp->size += b->size - sz; \
  b->size = sz; \
  FREE_MEM_INS_BLOCK(pp); \
  bl_uncommit(pp); \
}

static Block_t *bl_mapnew(size_t size)
{
  size_t map_size;
  Block_t *pp, *p;
  void *pt;
  
  map_size = PAGE_ALIGN(size);
  pt = (Block_t*)VirtualAlloc(NULL,map_size,MEM_RESERVE,PAGE_EXECUTE_READWRITE);
  if (pt == NULL) return (Block_t*)NULL;
  
  bl_last = pp = (Block_t*) bl_get();
  INIT_BLOCK(pp, (char*)pt, map_size);
  pp->broken = 1;
  
  return pp;
}

static void bl_commit(Block_t *b)
{
  char *u_start, *u_end;

  u_start = PAGE_DOWNALIGNP(b->ptr);
  u_end   = PAGE_ALIGNP(b->ptr+b->size);
  if (u_end <= u_start) return;

  VirtualAlloc((void*)u_start,u_end-u_start,MEM_COMMIT,PAGE_EXECUTE_READWRITE);
}

static void bl_uncommit(Block_t *b)
{
  char *u_start, *u_end;
  
  u_start = PAGE_ALIGNP(b->ptr);
  u_end   = PAGE_DOWNALIGNP(b->ptr+b->size);
  if (u_end <= u_start) return;

#if M_DOTRIMMING
  VirtualFree((void*)u_start,u_end-u_start,MEM_DECOMMIT);
#endif
}

/* requested size must be aligned to ALIGNMENT */
static Block_t *bl_alloc(size_t size)
{
  Block_t *p, *pp;
  
  /* try to find needed space in existing memory */
  for (p = free_mem_root, pp = NULL;p;)
  {
    if (p->size > size) { pp = p; p = p->l_free_mem; }
    else if (p->size < size) p = p->r_free_mem;
    else { pp = p; break; }
  }

  if (!pp)
  { /* map some memory - win32 can't grow memory blocks */
	if (size < LARGE_MPAGES*M_PAGESIZE)
		pp = bl_mapnew(LARGE_MPAGES*M_PAGESIZE);
	else
		pp = bl_mapnew(size + M_PAGESIZE);
  }
  
  /* just delete this node from free_mem tree */
  if (pp->next) free_mem_replace(pp->next); else free_mem_del(pp);
  pp->used = 1;
  
  if (pp->size - size > MALLOC_ALIGN)
  { /* this block can be splitted (it is unused,not_broken) */
    SPLIT_BLOCK(pp,size);
  }

  bl_commit(pp);
  
  return pp;
}

static void bl_free(Block_t *b)
{
  Block_t *p, *bl_next, *bl_prev;
  
  /* Look for blocks before & after `b` */
  for (p = ptrs_root, bl_next = NULL, bl_prev = NULL; p;)
  {
    if      (p->ptr > b->ptr) { bl_next = p; p = p->l_ptrs; }
    else if (p->ptr < b->ptr) { bl_prev = p; p = p->r_ptrs; }
    else break;
  }
  if (b->l_ptrs)
    for (bl_prev = b->l_ptrs; bl_prev->r_ptrs; bl_prev = bl_prev->r_ptrs);
  if (b->r_ptrs)
    for (bl_next = b->r_ptrs; bl_next->l_ptrs; bl_next = bl_next->l_ptrs);
  
  if (bl_next && !bl_next->broken && !bl_next->used)
  {
    FREE_MEM_DEL_BLOCK(bl_next)
    COMBINE_BLOCKS(b,bl_next)
  }

  if (bl_prev && !b->broken && !bl_prev->used)
  {
    FREE_MEM_DEL_BLOCK(bl_prev)
    COMBINE_BLOCKS(bl_prev,b)
    b = bl_prev;
  }
  
  b->used = 0;
  FREE_MEM_INS_BLOCK(b)
  bl_uncommit(b);
}

static void malloc_init(void)
{
  int i, mapsize, x, old_x, gcount;
  
  mapsize = M_PAGESIZE;
  
  __malloc_initialized = 0;
  bl_last = NULL;
  free_mem_root = NULL;
  ptrs_root = NULL;
  mapsize -= sizeof(Hunk_t);
  for (i = 1; i <= HUNK_MAXSIZE; i++)
  {
    free_h[i] = (Hunk_t*)NULL;
    for (x = mapsize/i, gcount = 0, old_x = 0; old_x != x;)
    {
      old_x = x;
      x = (mapsize - ALIGN(DIV8(old_x+7)))/i;
      if (gcount > 1 && x*i + ALIGN(DIV8(x+7)) <= mapsize) break;
      if (x*i + ALIGN(DIV8(x+7)) > mapsize) gcount++;
    }
    total_h[i] = x;
  }
  mutex_init(&malloc_lock);
  __malloc_initialized = 1;
}

/* PUBLIC functions */

void *mmalloc(size_t size)
{
  void *p;
  
  if (size == 0) return NULL;
 
  if (__malloc_initialized < 0) malloc_init();
  if (__malloc_initialized) mutex_lock(&malloc_lock);
  
  if (size <= HUNK_MAXSIZE)
    p = hunk_alloc(size);
  else
  {
    if ((p = bl_alloc(ALIGN(size))) != NULL)
      p = ((Block_t*)p)->ptr;
  }
  
  if (__malloc_initialized) mutex_unlock(&malloc_lock);
  
  return p;
}

void mfree(void *ptr)
{
  Block_t *p, *best;
 
  if (__malloc_initialized < 0) return;
  if (__malloc_initialized) mutex_lock(&malloc_lock);
  
  for (p = ptrs_root, best = NULL;p;)
  {
    if (p->ptr > (char*)ptr) p = p->l_ptrs;
    else { best = p; p = p->r_ptrs; }
  }
  
  if (!best || !best->used || best->ptr != (char*)ptr)
  {
    hunk_free((char*)ptr);
    if (__malloc_initialized) mutex_unlock(&malloc_lock);
    return;
  }
  
  bl_free(best);

  if (__malloc_initialized) mutex_unlock(&malloc_lock);
}

void *mrealloc_no_move(void *ptr, size_t size)
{
  Block_t *p, *best, *next;
  
  if (size <= HUNK_MAXSIZE) return NULL;
  
  if (__malloc_initialized <= 0) return mmalloc(size);
  
  mutex_lock(&malloc_lock);
  
  /* Locate block */
  for (p = ptrs_root, best = NULL;p;)
  {
    if (p->ptr > (char*)ptr) p = p->l_ptrs;
    else { best = p; p = p->r_ptrs; }
  }
  
  if (!best || !best->used || best->ptr != (char*)ptr)
  {
    mutex_unlock(&malloc_lock);
    return NULL;
  }
  
  size = ALIGN(size);
  
  if (size == best->size)
  {
    mutex_unlock(&malloc_lock);
    return ptr;
  }
  
  if (best->r_ptrs) /* get block just after */
    for (next = best->r_ptrs; next->l_ptrs; next = next->l_ptrs);
  else
    for (p = ptrs_root, next = NULL;p;)
    {
      if (p->ptr > best->ptr) { next = p; p = p->l_ptrs; }
      else if (p->ptr < best->ptr) p = p->r_ptrs;
      else break;
    }
  
  if (size < best->size)
  { /* shrink block */
    if (!next || next->used || next->broken)
    {
      if (best->size - size > MALLOC_ALIGN)
      { /* do split */
	SPLIT_BLOCK(best,size);
      }
    }
    else
    { /* just move border of next block */
      SHRINK_BLOCK(best,next,size);
    }
  }
  else if (next && !next->broken && !next->used)
  { /* can expand */
    if (best->size + next->size > size + HUNK_MAXSIZE)
    { /* shrink next free block */
      SHRINK_BLOCK(best,next,size);
    }
    else if (best->size + next->size >= size)
    { /* combine blocks (eat next one) */
      FREE_MEM_DEL_BLOCK(next);
      COMBINE_BLOCKS(best,next);
    }
    else
    { /* not enough memory in next block */
      mutex_unlock(&malloc_lock);
      return NULL;
    }
  }
  else
  { /* no next block */
    mutex_unlock(&malloc_lock);
    return NULL;
  }
  mutex_unlock(&malloc_lock);
  return best->ptr;
}

void *mrealloc(void *ptr, size_t size)
{
  void *tmp;
  
  tmp = mrealloc_no_move(ptr, size);
  
  if (!tmp)
  {
    Block_t *p, *best;
    
    mutex_lock(&malloc_lock);

    for (p = ptrs_root, best = NULL;p;)
    {
      if (p->ptr > (char*)ptr) p = p->l_ptrs;
      else { best = p; p = p->r_ptrs; }
    }
    
    if (!best || !best->used || best->ptr != (char*)ptr)
    {
      if (ptr)
      {
	Hunk_t *h;
        h = (Hunk_t*)PAGE_DOWNALIGNP(ptr);
        if (h->id == HUNK_ID)
        {
	  mutex_unlock(&malloc_lock);
	  if ((size >= HUNK_THRESHOLD && ALIGN(size) == h->size) ||
	      size == h->size) return ptr;
	  if ((tmp = mmalloc(size)) == NULL) return NULL;
	  mutex_lock(&malloc_lock);
	  memcpy(tmp,ptr,((size<h->size)?size:h->size));
	  hunk_free((char*)ptr);
	  mutex_unlock(&malloc_lock);
	  return tmp;
	}
      }
      mutex_unlock(&malloc_lock);
      return mmalloc(size);
    }
    
    mutex_unlock(&malloc_lock);
    
    /* copy whole block */
    if ((tmp = mmalloc(size)) == NULL) return NULL;
    memcpy(tmp,ptr,((size<best->size)?size:best->size));
    
    mutex_lock(&malloc_lock);
    bl_free(best);
    mutex_unlock(&malloc_lock);
  }
  return tmp;
}

void *mcalloc(size_t unit, size_t quantity)
{
  void *p;
  
  unit *= quantity;
 
  if ((p = mmalloc(unit)) == NULL) return NULL;
  memset(p,0,unit);
  return p;
}
